"""Writers module.

The witers module contains functions to construct the neccesary inputs for
other thermodynamics libraries.

Supported:

- Clapeyron.jl: https://github.com/ClapeyronThermo/Clapeyron.jl
- Caleb's Bell thermo: https://github.com/CalebBell/thermo
"""

from ugropy.writers.clapeyron import to_clapeyron
from ugropy.writers.thermo import to_thermo
from ugropy.writers.yaeosf import to_yaeos


__all__ = ["to_clapeyron", "to_thermo", "to_yaeos"]



"""Abdulelah-Gani frag classes module."""

from . import (
    abdulelah_gani,
    abdulelah_gani_pst,
    abdulelah_gani_pst_result,
    abdulelah_gani_result,
)


__all__ = [
    "abdulelah_gani_pst",
    "abdulelah_gani_pst_result",
    "abdulelah_gani_result",
    "abdulelah_gani",
]



"""Molar mass writer module."""

import pathlib
from io import StringIO
from typing import List

import numpy as np

import pandas as pd

from ugropy.models.dortmundmod import dortmund
from ugropy.models.psrkmod import psrk
from ugropy.models.unifacmod import unifac


def write_molar_mass(
    path: pathlib.Path,
    batch_name: str,
    molecules_names: List[str],
    unifac_groups: List[dict] = [],
    psrk_groups: List[dict] = [],
    dortmund_groups: List[dict] = [],
    property_estimator: List = [],
) -> None:
    """Create the DataFrame with the molecular weights for Clapeyron.jl.

    Parameters
    ----------
    path : pathlib.Path
        Path to the directory to store de .csv files, by default "./database".
    batch_name : str, optional
        Name of the writing batch. For example, if you name the batch with
        "batch1", the output of the UNIFAC groups will be:
        "batch1_ogUNIFAC_groups.csv". With the default value will be
        "ogUNIFAC_groups.csv", by default "".
    molecules_names : List[str]
        List of names for each chemical to write in the .csv files.
    unifac_groups : List[dict], optional
        List of classic liquid-vapor UNIFAC groups, by default [].
    psrk_groups : List[dict], optional
        List of Predictive Soave-Redlich-Kwong groups, by default [].
    dortmund_groups : List[dict], optional
        List of Dortmund UNIFAC groups, by default [].
    property_estimator : List, optional
        List of JobackFragmentationResult or AGaniFragmentationResult, by
        default [].
    """
    data_str = (
        "Clapeyron Database File,,\n"
        "Molar Mases Single Params,,\n"
        "species,CAS,Mw\n"
    )
    path_molar_mass = pathlib.Path(path)
    # =========================================================================
    # Get molecular weights
    # =========================================================================
    if property_estimator:
        molecular_weigths = [
            p.molecular_weight.to("g/mol").magnitude
            for p in property_estimator
        ]
    elif unifac_groups:
        df = unifac.subgroups.copy()
        molecular_weigths = []
        for groups in unifac_groups:
            contribution = df.loc[groups.keys(), "molecular_weight"].to_numpy()
            molecular_weigths.append(
                np.dot(contribution, list(groups.values()))
            )
    elif psrk_groups:
        df = psrk.subgroups.copy()
        molecular_weigths = []
        for groups in psrk_groups:
            contribution = df.loc[groups.keys(), "molecular_weight"].to_numpy()
            molecular_weigths.append(
                np.dot(contribution, list(groups.values()))
            )
    elif dortmund_groups:
        df = dortmund.subgroups.copy()
        molecular_weigths = []
        for groups in dortmund_groups:
            contribution = df.loc[groups.keys(), "molecular_weight"].to_numpy()
            molecular_weigths.append(
                np.dot(contribution, list(groups.values()))
            )
    else:
        raise ValueError(
            "Joback, UNIFAC, PSRK or Dortmund groups must be provided."
        )

    # =========================================================================
    # Build dataframe
    # =========================================================================
    df = pd.read_csv(StringIO(data_str))

    for idx, name in enumerate(molecules_names):
        new_row = {
            "Clapeyron Database File": name,
            "Unnamed: 1": "",
            "Unnamed: 2": molecular_weigths[idx],
        }
        df.loc[len(df)] = new_row

    df.columns = ["" if "Unnamed" in col else col for col in df.columns]

    if batch_name == "":
        with open(
            path_molar_mass / "molarmass.csv",
            "w",
            newline="",
            encoding="utf-8",
        ) as file:
            df.to_csv(file, index=False)

    else:
        with open(
            path_molar_mass / f"{batch_name}_molarmass.csv",
            "w",
            newline="",
            encoding="utf-8",
        ) as file:
            df.to_csv(file, index=False)



"""Critical properties writer module for Clapeyron.jl."""

import pathlib
from io import StringIO
from typing import List

import pandas as pd


def write_critical(
    path: pathlib.Path,
    batch_name: str,
    molecules_names: List[str],
    property_estimator: List = [],
) -> None:
    """Create the DataFrame with the critical properties for Clapeyron.jl.

    Uses the estimated critical properties from any property estimator of
    ugropy.

    Parameters
    ----------
    path : pathlib.Path, optional
        Path to the directory to store de .csv files, by default "./database".
    batch_name : str, optional
        Name of the writing batch. For example, if you name the batch with
        "batch1", the output of the UNIFAC groups will be:
        "batch1_ogUNIFAC_groups.csv". With the default value will be
        "ogUNIFAC_groups.csv", by default "".
    molecules_names : List[str]
        List of names for each chemical to write in the .csv files.
    property_estimator : List, optional
        List of JobackFragmentationResult or AGaniFragmentationResult, by
        default [].
    """
    data_str = (
        "Clapeyron Database File,,,,,\n"
        "Critical Single Parameters,,,,,\n"
        "species,CAS,Tc,Pc,Vc,acentricfactor\n"
    )
    path_critical = pathlib.Path(path)
    # =========================================================================
    # Build dataframe
    # =========================================================================
    df = pd.read_csv(StringIO(data_str))

    for idx, name in enumerate(molecules_names):
        new_row = {
            "Clapeyron Database File": name,
            "Unnamed: 1": "",
            "Unnamed: 2": property_estimator[idx]
            .critical_temperature.to("K")
            .magnitude,
            "Unnamed: 3": property_estimator[idx]
            .critical_pressure.to("Pa")
            .magnitude,
            "Unnamed: 4": property_estimator[idx]
            .critical_volume.to("m^3/mol")
            .magnitude,
            "Unnamed: 5": property_estimator[idx].acentric_factor.magnitude,
        }
        df.loc[len(df)] = new_row

    df.columns = ["" if "Unnamed" in col else col for col in df.columns]

    if batch_name == "":
        with open(
            path_critical / "critical.csv", "w", newline="", encoding="utf-8"
        ) as file:
            df.to_csv(file, index=False)

    else:
        with open(
            path_critical / f"{batch_name}_critical.csv",
            "w",
            newline="",
            encoding="utf-8",
        ) as file:
            df.to_csv(file, index=False)



"""UNIFAC groups writer module."""

import pathlib
from typing import List


def write_unifac(
    path: pathlib.Path,
    batch_name: str,
    molecules_names: List[str],
    unifac_groups: List[dict],
) -> None:
    """Create the DataFrame with the classic LV-UNIFAC groups for Clapeyron.jl.

    Parameters
    ----------
    path : pathlib.Path
        Path to the directory to store de .csv files, by default "./database".
    batch_name : str, optional
        Name of the writing batch. For example, if you name the batch with
        "batch1", the output of the UNIFAC groups will be:
        "batch1_ogUNIFAC_groups.csv". With the default value will be
        "ogUNIFAC_groups.csv", by default "".
    molecules_names : List[str]
        List of names for each chemical to write in the .csv files.
    unifac_groups : List[dict], optional
        List of classic liquid-vapor UNIFAC groups.
    """
    lines = [
        "Clapeyron Database File,\n"
        "original UNIFAC Groups,[csvtype = groups,grouptype = originalUNIFAC]\n"  # noqa
        "species,groups\n"
    ]

    path_ogunifac = path / "ogUNIFAC"

    for name, groups in zip(molecules_names, unifac_groups):
        groups_str = '"['

        for group in groups.keys():
            groups_str += f'""{group}"" => {groups[group]}, '

        groups_str = groups_str[: len(groups_str) - 2]
        groups_str += ']"\n'

        new_line = [f"{name},{groups_str}"]

        lines.extend(new_line)

    # Create folder for ogUNIFAC groups
    if not path_ogunifac.is_dir():
        path_ogunifac.mkdir(parents=True)

    # Write .csv
    if batch_name == "":
        write_path = path_ogunifac / "ogUNIFAC_groups.csv"
    else:
        write_path = path_ogunifac / f"{batch_name}_ogUNIFAC_groups.csv"

    with open(write_path, "w", encoding="utf-8", newline="\n") as file:
        file.writelines(lines)



"""Abdulelah-Gani fragmentation model result module :cite:p:`gani`."""

import numpy as np

import pandas as pd

import pint

from rdkit import Chem
from rdkit.Chem import Descriptors

from ugropy.constants import ureg
from ugropy.core.frag_classes.abdulelah_gani.abdulelah_gani_pst_result import (
    AGaniPSTFragmentationResult,
)


class AGaniFragmentationResult:
    """Abdulelah-Gani fragmentation model result :cite:p:`gani`.

    Class to store the results of the Abdulelah-Gani fragmentation model.

    Parameters
    ----------
    molecule : Chem.rdchem.Mol
        RDKit molecule object.
    primary_fragmentation : AGaniPSTFragmentationResult
        Primary fragmentation model result.
    secondary_fragmentation : AGaniPSTFragmentationResult
        Secondary fragmentation model result.
    tertiary_fragmentation : AGaniPSTFragmentationResult
        Tertiary fragmentation model result.
    properties_contributions : pd.DataFrame
        Contribution parameters of each group for each property of the model.
    properties_biases : pd.DataFrame
        Biases parameters of each property of the model.

    Attributes
    ----------
    molecule : Chem.rdchem.Mol
        RDKit molecule object.
    primary : AGaniPSTFragmentationResult
        Primary fragmentation model result.
    secondary : AGaniPSTFragmentationResult
        Secondary fragmentation model result.
    tertiary : AGaniPSTFragmentationResult
        Tertiary fragmentation model result.
    ml_vector : np.ndarray
        Vector of groups occurrences to evaluate ML model.
    molecular_weight : pint.Quantity
        Molecular weight [g/mol].
    critical_temperature : pint.Quantity
        Critical temperature [K] GC-Simple method.
    critical_pressure : pint.Quantity
        Critical pressure [bar] GC-Simple method.
    critical_volume : pint.Quantity
        Critical volume [cm³/mol] GC-Simple method.
    acentric_factor : pint.Quantity
        Acentric factor [-] GC-Simple method.
    liquid_molar_volume : pint.Quantity
        Liquid molar volume [L/mol] GC-Simple method.
    ig_formation_enthalpy : pint.Quantity
        Ideal gas formation enthalpy [kJ/mol] GC-Simple method.
    ig_formation_gibbs : pint.Quantity
        Ideal gas formation Gibbs [kJ/mol] GC-Simple method.
    """

    def __init__(
        self,
        molecule: Chem.rdchem.Mol,
        primary_fragmentation: AGaniPSTFragmentationResult,
        secondary_fragmentation: AGaniPSTFragmentationResult,
        tertiary_fragmentation: AGaniPSTFragmentationResult,
        properties_contributions: pd.DataFrame,
        properties_biases: pd.DataFrame,
    ) -> None:

        # =====================================================================
        # Fragmentation models
        # =====================================================================
        self.molecule = molecule
        self.primary = primary_fragmentation
        self.secondary = secondary_fragmentation
        self.tertiary = tertiary_fragmentation

        # =====================================================================
        # ML Vector. Vector of parameters occurrences to evaluate ML model
        # =====================================================================
        self.ml_vector = np.zeros(424, dtype=np.int64)

        for group, ocurrences in self.primary.subgroups_numbers.items():
            self.ml_vector[group - 1] = ocurrences

        for group, ocurrences in self.secondary.subgroups_numbers.items():
            self.ml_vector[group - 1] = ocurrences

        for group, ocurrences in self.tertiary.subgroups_numbers.items():
            self.ml_vector[group - 1] = ocurrences

        self.ml_vector = self.ml_vector.reshape(1, -1)

        # =====================================================================
        # Properties
        # =====================================================================
        self.molecular_weight: pint.Quantity = (
            Descriptors.MolWt(self.molecule) * ureg.g / ureg.mol
        )
        self.critical_temperature: pint.Quantity = None
        self.critical_pressure: pint.Quantity = None
        self.critical_volume: pint.Quantity = None
        self.acentric_factor: pint.Quantity = None
        self.liquid_molar_volume: pint.Quantity = None
        self.ig_formation_enthalpy: pint.Quantity = None
        self.ig_formation_gibbs: pint.Quantity = None

        if self.primary.subgroups != {}:
            self.properties_calculation(
                properties_contributions, properties_biases
            )

    def properties_calculation(
        self,
        properties_contributions: pd.DataFrame,
        properties_biases: pd.DataFrame,
    ) -> None:
        """Calculate the properties with the fragmentation results.

        Properties are calculated with the GC-Simple method.

        Parameters
        ----------
        properties_contributions : pd.DataFrame
            Contribution parameters of each group for each property of the
            model.
        properties_biases : pd.DataFrame
            Biases parameters of each property of the model.
        """
        # Critical temperature
        tc_c = properties_contributions["Tc"].values
        tc_b = properties_biases["Tc"].loc["b1"]

        tc_sum = np.dot(tc_c, self.ml_vector.T)[0]

        if tc_sum <= 0:
            self.critical_temperature = np.nan * ureg.K
        else:
            self.critical_temperature = tc_b * np.log(tc_sum) * ureg.K

        # Critical pressure
        pc_c = properties_contributions["Pc"].values
        pc_b1 = properties_biases["Pc"].loc["b1"]
        pc_b2 = properties_biases["Pc"].loc["b2"]

        pc_sum = np.dot(pc_c, self.ml_vector.T)[0]

        self.critical_pressure = (
            (pc_sum + pc_b2) ** (-1 / 0.5) + pc_b1
        ) * ureg.bar

        # Critical volume
        vc_c = properties_contributions["Vc"].values
        vc_b1 = properties_biases["Vc"].loc["b1"]

        vc_sum = np.dot(vc_c, self.ml_vector.T)[0]

        self.critical_volume = (vc_sum + vc_b1) * ureg.cm**3 / ureg.mol

        # Acentric factor
        w = properties_contributions["w"].values
        w_b1 = properties_biases["w"].loc["b1"]
        w_b2 = properties_biases["w"].loc["b2"]
        w_b3 = properties_biases["w"].loc["b3"]

        w_sum = np.dot(w, self.ml_vector.T)[0]

        self.acentric_factor = (
            np.log((w_sum + w_b3) ** (1 / w_b2)) * w_b1 * ureg.dimensionless
        )

        # Liquid molar volume
        lmv_c = properties_contributions["Lmv"].values
        lmv_b1 = properties_biases["Lmv"].loc["b1"]

        lmv_sum = np.dot(lmv_c, self.ml_vector.T)[0]

        self.liquid_molar_volume = (lmv_sum + lmv_b1) * ureg.L / ureg.mol

        # Ideal gas formation enthalpy
        h_f_c = properties_contributions["Hf"].values
        h_f_c_b1 = properties_biases["Hf"].loc["b1"]

        h_f_sum = np.dot(h_f_c, self.ml_vector.T)[0]

        self.ig_formation_enthalpy = (h_f_sum + h_f_c_b1) * ureg.kJ / ureg.mol

        # Ideal gas formation Gibbs
        g_f_c = properties_contributions["Gf"].values
        g_f_c_b1 = properties_biases["Gf"].loc["b1"]

        g_f_sum = np.dot(g_f_c, self.ml_vector.T)[0]

        self.ig_formation_gibbs = (g_f_sum + g_f_c_b1) * ureg.kJ / ureg.mol



"""Dortmund mod-UNIFAC FragmentationModel implementation.

Import and use the Dortmund FragmentationModel with:

.. code-block:: python

    from ugropy import dortmund

    # Get groups from molecule's name
    tol = dortmund.get_groups("toluene")

    print(tol.subgroups)

    # Get groups from molecule's SMILES
    eth = dortmund.get_groups("CCO", "smiles")

    print(eth.subgroups)

Attributes
----------
dortmund : GibbsModel
    Dortmund mod-UNIFAC FragmentationModel :cite:p:`dor1, dor2, dor3, dor4,
    dor5, dor6, dor7, dor8, dor9, dor10, dor11`
"""

from ugropy.constants import _csvs
from ugropy.core.frag_classes.gibbs_model.gibbs_model import GibbsModel
from ugropy.models.read_csv import _rd


# =============================================================================
# Dortmund UNIFAC
# =============================================================================
_dor = _csvs / "dortmund"

_dor_sg = _rd(_dor / "dortmund_subgroups.csv", "group")
_dor_info = _rd(_dor / "dortmund_info.csv", "group")

dortmund = GibbsModel(
    subgroups=_dor_sg, subgroups_info=_dor_info, calculate_r_q=False
)



"""Groups class module."""

from typing import List, Union

from rdkit.Chem import Descriptors

from .core.frag_classes.abdulelah_gani.abdulelah_gani_result import (
    AGaniFragmentationResult,
)
from .core.frag_classes.gibbs_model.gibbs_result import (
    GibbsFragmentationResult,
)
from .core.frag_classes.joback.joback_result import JobackFragmentationResult
from .core.get_rdkit_object import instantiate_mol_object
from .core.ilp_solvers.default_solver import DefaultSolver
from .core.ilp_solvers.ilp_solver import ILPSolver
from .models.abdulelah_gani_mod import abdulelah_gani
from .models.dortmundmod import dortmund
from .models.jobackmod import joback
from .models.psrkmod import psrk
from .models.unifacmod import unifac


class Groups:
    """Groups class.

    Stores the solved FragmentationModels results of a molecule. This class
    was implemented on an early version of the `ugropy` library. Is not really
    meant to be used, instead is recommended to use directly the corresponding
    FragmentationModel independently since it provides more flexibility and
    control over the results. The class is kept since in most cases is very
    comfortable to have all the results in a single object.

    Parameters
    ----------
    identifier : Union[str, Chem.rdchem.Mol]
        Identifier of the molecule. You can use either the name of the
        molecule, the SMILEs of the molecule or a rdkit Mol object.
    identifier_type : str, optional
        Identifier type of the molecule. Use "name" if you are providing
        the molecules' name, "smiles" if you are providing the SMILES
        or "mol" if you are providing a rdkir mol object, by default "name"
    solver : ILPSolver, optional
        ILP solver class, by default DefaultSolver
    search_multiple_solutions : bool, optional
        Weather search for multiple solutions or not, by default False
        If False the return will be a FragmentationResult object, if True
        the return will be a list of FragmentationResult objects.
    search_nonoptimal : bool, optional
        If True, the solver will search for non-optimal solutions along with
        the optimal ones. This is useful when the user wants to find all
        possible combinations of fragments that cover the universe. By default
        False. If `search_multiple_solutions` is False, this parameter will be
        ignored.
    normal_boiling_temperature : float, optional
        If provided, will be used to estimate critical temperature, acentric
        factor, and vapor pressure instead of the estimated normal boiling
        point in the Joback group contribution model, by default None.

    Attributes
    ----------
    identifier : str
        Identifier of a molecule. Example: 'hexane' or 'CCCCCC'.
    identifier_type : str, optional
        Use 'name' to search a molecule by name or 'smiles' to provide the
        molecule SMILES representation, by default "name".
    mol_object : rdkit.Chem.rdchem.Mol
        RDKit Mol object.
    molecular_weight : float
        Molecule's molecular weight from rdkit.Chem.Descriptors.MolWt [g/mol].
    unifac : Union[GibbsFragmentationResult, List[GibbsFragmentationResult]]
        Classic LV-UNIFAC subgroups.
    psrk : Union[GibbsFragmentationResult, List[GibbsFragmentationResult]]
        Predictive Soave-Redlich-Kwong subgroups.
    dortmund : Union[GibbsFragmentationResult, List[GibbsFragmentationResult]]
        Dortmund UNIFAC subgroups.
    joback : Union[JobackFragmentationResult, List[JobackFragmentationResult]]
        JobackFragmentationResult object that contains the Joback subgroups and
        the estimated properties of the molecule.
    agani : Union[AGaniFragmentationResult, List[AGaniFragmentationResult]]
        AGaniFragmentationResult object that contains the Abdulelah-Gani
        subgroups and the estimated properties of the molecule.
    """

    def __init__(
        self,
        identifier: str,
        identifier_type: str = "name",
        solver: ILPSolver = DefaultSolver,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
        normal_boiling_temperature: float = None,
    ) -> None:
        self.identifier_type = identifier_type.lower()
        self.identifier = identifier
        self.mol_object = instantiate_mol_object(identifier, identifier_type)
        self.molecular_weight = Descriptors.MolWt(self.mol_object)

        # UNIFAC
        self.unifac: Union[
            GibbsFragmentationResult, List[GibbsFragmentationResult]
        ] = unifac.get_groups(
            self.identifier,
            self.identifier_type,
            solver=solver,
            search_multiple_solutions=search_multiple_solutions,
            search_nonoptimal=search_nonoptimal,
        )

        # PSRK
        self.psrk: Union[
            GibbsFragmentationResult, List[GibbsFragmentationResult]
        ] = psrk.get_groups(
            self.identifier,
            self.identifier_type,
            solver=solver,
            search_multiple_solutions=search_multiple_solutions,
            search_nonoptimal=search_nonoptimal,
        )

        # Dortmund
        self.dortmund: Union[
            GibbsFragmentationResult, List[GibbsFragmentationResult]
        ] = dortmund.get_groups(
            self.identifier,
            self.identifier_type,
            solver=solver,
            search_multiple_solutions=search_multiple_solutions,
            search_nonoptimal=search_nonoptimal,
        )

        # Joback
        self.joback: Union[
            JobackFragmentationResult, List[JobackFragmentationResult]
        ] = joback.get_groups(
            self.identifier,
            self.identifier_type,
            solver=solver,
            search_multiple_solutions=search_multiple_solutions,
            normal_boiling_point=normal_boiling_temperature,
            search_nonoptimal=search_nonoptimal,
        )

        # Abdulelah-Gani
        self.agani: Union[
            AGaniFragmentationResult, List[AGaniFragmentationResult]
        ] = abdulelah_gani.get_groups(
            self.identifier,
            self.identifier_type,
            solver=solver,
            search_multiple_solutions=search_multiple_solutions,
            search_nonoptimal=search_nonoptimal,
        )



"""Joback fragmentation module."""

from typing import List, Union

import pandas as pd

from rdkit import Chem

from ugropy.core.frag_classes.abdulelah_gani.abdulelah_gani_pst_result import (
    AGaniPSTFragmentationResult,
)
from ugropy.core.frag_classes.base.fragmentation_model import (
    FragmentationModel,
)
from ugropy.core.ilp_solvers.default_solver import DefaultSolver
from ugropy.core.ilp_solvers.ilp_solver import ILPSolver


class AbdulelahGaniPSTModel(FragmentationModel):
    """Abdulelah-Gani model dedicated to properties estimation.

    Class to construct the primary, secondary and tertiary structures detector
    for the Abdulelah-Gani properties estimation model :cite:p:`gani`.

    Parameters
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Mandatory columns:
        'smarts' (SMARTS representations of the group to detect its precense in
        the molecule).
    subgroups_info : pd.DataFrame
        Group's subgroups numbers.

    Attributes
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Mandatory columns:
        'smarts' (SMARTS representations of the group to detect its precense in
        the molecule).
    detection_mols : dict
        Dictionary cotaining all the rdkit Mol object from the detection_smarts
        subgroups column
    info : pd.DataFrame
        Group's subgroups numbers.
    """

    def __init__(
        self,
        subgroups: pd.DataFrame,
        subgroups_info: pd.DataFrame,
        allow_overlapping: bool = False,
        allow_free_atoms: bool = False,
    ) -> None:

        super().__init__(
            subgroups=subgroups,
            allow_overlapping=allow_overlapping,
            allow_free_atoms=allow_free_atoms,
            fragmentation_result=AGaniPSTFragmentationResult,
        )

        self.subgroups_info = subgroups_info

    def get_groups(
        self,
        identifier: Union[str, Chem.rdchem.Mol],
        identifier_type: str = "name",
        solver: ILPSolver = DefaultSolver,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
    ) -> Union[AGaniPSTFragmentationResult, List[AGaniPSTFragmentationResult]]:
        """Get the groups of a molecule.

        Parameters
        ----------
        identifier : Union[str, Chem.rdchem.Mol]
            Identifier of the molecule. You can use either the name of the
            molecule, the SMILEs of the molecule or a rdkit Mol object.
        identifier_type : str, optional
            Identifier type of the molecule. Use "name" if you are providing
            the molecules' name, "smiles" if you are providing the SMILES or
            "mol" if you are providing a rdkir mol object, by default "name"
        solver : ILPSolver, optional
            ILP solver class, by default DefaultSolver
        search_multiple_solutions : bool, optional
            Weather search for multiple solutions or not, by default False If
            False the return will be a FragmentationResult object, if True the
            return will be a list of FragmentationResult objects.
        search_nonoptimal : bool, optional
            If True, the solver will search for non-optimal solutions along
            with the optimal ones. This is useful when the user wants to find
            all possible combinations of fragments that cover the universe. By
            default False. If `search_multiple_solutions` is False, this
            parameter will be ignored.

        Returns
        -------
        Union[AGaniPFragmentationResult, List[AGaniPFragmentationResult]]
            Fragmentation result. If search_multiple_solutions is False the
            return will be a FragmentationResult object, if True the return
            will be a list of FragmentationResult objects.
        """
        sol = super().get_groups(
            identifier,
            identifier_type,
            solver,
            search_multiple_solutions,
            search_nonoptimal=search_nonoptimal,
            subgroups_info=self.subgroups_info,
        )

        return sol

    def mol_preprocess(self, mol: Chem.rdchem.Mol) -> Chem.rdchem.Mol:
        """Preprocess the molecule to be ready for the fragmentation.

        This method preprocess the molecule to be ready for the fragmentation
        process. The preprocessing steps are:

        1. Kekulize the molecule.
        2. Identify the aromatic rings.
        3. Check the aromaticity of the rings.
        4. Make the rings aromatic or non-aromatic based on the setp 3 check.

        The criteria to check the aromaticity of the rings are based on the
        criteria proposed by Abdulelah-Gani in the original paper datase
        :cite:p:`gani`.

        Parameters
        ----------
        mol : Chem.rdchem.Mol
            Molecule to preprocess

        Returns
        -------
        Chem.rdchem.Mol
            Preprocessed molecule
        """
        # Clone of the molecule to kekulize
        kekulized_mol = Chem.Mol(mol)
        Chem.Kekulize(kekulized_mol, clearAromaticFlags=True)

        # Step 1: Identify all ring on the molecule
        ring_info = mol.GetRingInfo()
        atom_rings = ring_info.AtomRings()

        # Step 2: Get only the aromatic rings
        rdkit_aromatic_rings = [
            ring
            for ring in atom_rings
            if all(mol.GetAtomWithIdx(idx).GetIsAromatic() for idx in ring)
        ]

        # Lists to store the AGani aromatic rings and not aromatic rings
        aromatic_rings = []
        non_aromatic_rings = []

        # Step 3: Check which rings are aromatic
        for ring in rdkit_aromatic_rings:
            if self.check_ring_aromaticity(mol, ring):
                aromatic_rings.append(ring)
            else:
                non_aromatic_rings.append(ring)

        # Step 4: Preprocess the non aromatic rings
        for ring in non_aromatic_rings:
            # Make atoms of the ring non-aromatic
            for idx in ring:
                mol.GetAtomWithIdx(idx).SetIsAromatic(False)

            # Make bonds of the ring non-aromatic
            for i in range(len(ring)):
                atom1, atom2 = ring[i], ring[(i + 1) % len(ring)]
                bond = mol.GetBondBetweenAtoms(atom1, atom2)

                # Set bond type to the kekulized bond type
                kekulized_bond = kekulized_mol.GetBondBetweenAtoms(
                    atom1, atom2
                )
                bond.SetBondType(kekulized_bond.GetBondType())

        # Step 5: Revisit aromatic rings to make sure they are aromatic
        for ring in aromatic_rings:
            # Restore atoms aromaticity
            for idx in ring:
                mol.GetAtomWithIdx(idx).SetIsAromatic(True)

            # Restore bonds aromaticity
            for i in range(len(ring)):
                atom1, atom2 = ring[i], ring[(i + 1) % len(ring)]
                bond = mol.GetBondBetweenAtoms(atom1, atom2)
                bond.SetBondType(Chem.rdchem.BondType.AROMATIC)

        return mol

    def check_ring_aromaticity(
        self, mol: Chem.rdchem.Mol, ring: List[int]
    ) -> bool:
        """Verify ring aromaticity along with the authors criteria.

        Parameters
        ----------
        mol : Chem.Mol
            Molecule
        ring : list[int]
            List of atom indexes that form the ring

        Returns
        -------
        bool
            True if the ring is aromatic, False otherwise
        """
        # Non-aromatic patterns
        non_aromatic_patterns = [
            # "[nH0]1[cH0][nH0][cH0][cH0][cH0]1",
            # "[nH0]1[cH0][nH0][cH0][nH0][cH0]1",
            # "[nH0X2]1[c;R2][c;R2][nH0X3][c;R2][c;R2]1",
            "n1[cH0;$([cH0]=O)]nccc1",
            "n1ccnc[cH0;$([cH0]=O)]1",
            "n1ccn[cH0;$([cH0]=O)][cH0;$([cH0]=O)]1",
            "[nH0]1[cH0;$([cH0]=O)][nH0]cn[cH0;$([cH0]=O)]1",
            "c1ccco1",
            "n1[nH0][cH0;$([cH0]=O)]ccc1",
            "[cH0]1[cH0]cc[cH0]cc1",
            "[nH0;R2]1ccccc1",
            "n1[cH0;$([cH0]=*)]cccc1",
            "n1c[cH0;$([cH0]=*)]ccc1",
            "n1cc[cH0;$([cH0]=*)]cc1",
            "n1ccc[cH0;$([cH0]=*)]c1",
            "n1cccc[cH0;$([cH0]=*)]1",
            "n1cn[cH0;$([cH0]=O)]cc1",
        ]

        # Aromatic patterns
        aromatic_patterns = [
            "c1ccccc1",
            "c1ncncn1",
            "n1ccccc1",
            "n1ccncc1",
            "[nH0;R1]1[c;R1][nH0]ccc1",
            "[nH0]1[nH0]cccc1",
            "n1ccccc1",
            "[nX2]1[nX2]c[nX2]cc1",
            "[nH0]1c[nH0]ccc1",
        ]

        # Ring as set to compare
        ring_set = set(ring)

        # Check if the ring matches any non aromatic pattern
        for smarts in non_aromatic_patterns:
            pattern = Chem.MolFromSmarts(smarts)

            for match in mol.GetSubstructMatches(pattern):
                match_set = set(match)
                if match_set == ring_set:
                    return False

        # Check if the ring matches any aromatic pattern
        for smarts in aromatic_patterns:
            pattern = Chem.MolFromSmarts(smarts)

            for match in mol.GetSubstructMatches(pattern):
                match_set = set(match)
                if match_set == ring_set:
                    return True

        # Special case: all are carbon atoms
        if all(mol.GetAtomWithIdx(idx).GetSymbol() == "C" for idx in ring):
            return True

        # Default case, we consider the ring as non aromatic
        return False



"""to_clapeyron module."""

import pathlib
from typing import List

from ugropy.writers.clapeyron_writers.critical import write_critical
from ugropy.writers.clapeyron_writers.dortmund_groups import write_dortmund
from ugropy.writers.clapeyron_writers.molar_mass import write_molar_mass
from ugropy.writers.clapeyron_writers.psrk_groups import write_psrk
from ugropy.writers.clapeyron_writers.unifac_groups import write_unifac


def to_clapeyron(
    molecules_names: List[str],
    unifac_groups: List[dict] = [],
    psrk_groups: List[dict] = [],
    dortmund_groups: List[dict] = [],
    property_estimator: List = [],
    path: str = "database",
    batch_name: str = "",
) -> None:
    """Write the .csv input files for Clapeyron.jl.

    The provided lists must have the same length. If one of the model lists is
    left empty, that model will not be writen in the .csv.

    Parameters
    ----------
    molecules_names : List[str]
        List of names for each chemical to write in the .csv files.
    unifac_groups : List[dict], optional
        List of classic liquid-vapor UNIFAC groups, by default [].
    psrk_groups : List[dict], optional
        List of Predictive Soave-Redlich-Kwong groups, by default [].
    dortmund_groups : List[dict], optional
        List of Dortmund UNIFAC groups, by default [].
    property_estimator : List, optional
        List of JobackFragmentationResult or AGaniFragmentationResult, by
        default [].
    path : str, optional
        Path to the directory to store de .csv files, by default "./database".
    batch_name : str, optional
        Name of the writing batch. For example, if you name the batch with
        "batch1", the output of the UNIFAC groups will be:
        "batch1_ogUNIFAC_groups.csv". With the default value will be
        "ogUNIFAC_groups.csv", by default "".
    """
    # Use pathlib's Path internally
    path_pathlib = pathlib.Path(path)

    # Check if all list have correct data:
    if len(molecules_names) == 0:
        raise ValueError("No names provided for the molecules.")

    if unifac_groups and len(unifac_groups) != len(molecules_names):
        raise ValueError(
            "UNIFAC groups list must have the same amount of elements than"
            "the molecules name list."
        )

    if psrk_groups and len(psrk_groups) != len(molecules_names):
        raise ValueError(
            "PSRK groups list must have the same amount of elements than"
            "the molecules name list."
        )

    if dortmund_groups and len(dortmund_groups) != len(molecules_names):
        raise ValueError(
            "Dortmund groups list must have the same amount of elements than"
            "the molecules name list."
        )

    if property_estimator and len(property_estimator) != len(molecules_names):
        raise ValueError(
            "Property estimators result objects list must have the same amount"
            "of elements than the molecules name list."
        )

    # Create dir if not created
    if not path_pathlib.is_dir():
        path_pathlib.mkdir(parents=True)

    # Molar mass
    write_molar_mass(
        path_pathlib,
        batch_name,
        molecules_names,
        unifac_groups,
        psrk_groups,
        dortmund_groups,
        property_estimator,
    )

    # LV-UNIFAC
    if unifac_groups:
        write_unifac(path_pathlib, batch_name, molecules_names, unifac_groups)

    # PSRK
    if psrk_groups:
        write_psrk(path_pathlib, batch_name, molecules_names, psrk_groups)

    # Dortmund
    if dortmund_groups:
        write_dortmund(
            path_pathlib, batch_name, molecules_names, dortmund_groups
        )

    # Critical
    if property_estimator:
        write_critical(
            path_pathlib, batch_name, molecules_names, property_estimator
        )



"""Joback FragmentationModel implementation.

Import and use the Joback FragmentationModel with:

.. code-block:: python

    from ugropy import joback

    # Get groups from molecule's name
    tol = joback.get_groups("toluene")

    print(tol.subgroups)

    # Get groups from molecule's SMILES
    eth = joback.get_groups("CCO", "smiles")

    print(eth.subgroups)

Attributes
----------
joback: JobackModel
    Joback FragmentationModel :cite:p:`joback1, joback2`
"""

from ugropy.constants import _csvs
from ugropy.core.frag_classes.joback.joback_model import JobackModel
from ugropy.models.read_csv import _rd

# =============================================================================
# Joback
# =============================================================================
_jb = _csvs / "joback"

_jb_sg = _rd(_jb / "joback_subgroups.csv", "group")
_jb_props = _rd(_jb / "properties_contrib.csv", "group")

joback = JobackModel(subgroups=_jb_sg, properties_contributions=_jb_props)



"""Abdulelah-Gani FragmentationModel implementation.

Import and use with:

.. code-block:: python

    from ugropy import abdulelah_gani

    # Get groups from molecule's name
    tol = abdulelah_gani.get_groups("toluene")

    print(tol.critical_temperature)
    print(tol.primary.subgroups)
    print(tol.secondary.subgroups)
    print(tol.tertiary.subgroups)

    # Get groups from molecule's SMILES
    eth = abdulelah_gani.get_groups("CCO", "smiles")

    print(eth.critical_pressure)

Attributes
----------
abdulelah_gani: AbdulelahGaniModel
    Abdulelah-Gani FragmentationModel :cite:p:`gani`
"""

from ugropy.constants import _csvs
from ugropy.core.frag_classes.abdulelah_gani.abdulelah_gani import (
    AbdulelahGaniModel,
)
from ugropy.models.abdulelah_gani_pmod import abdulelah_gani_p
from ugropy.models.abdulelah_gani_smod import abdulelah_gani_s
from ugropy.models.abdulelah_gani_tmod import abdulelah_gani_t
from ugropy.models.read_csv import _rd

# =============================================================================
# Abdulelah Gani Model
# =============================================================================
_ag = _csvs / "abdulelah_gani"

_ag_pc = _rd(_ag / "properties_contributions.csv", "group")
_ag_biases = _rd(_ag / "properties_biases.csv", "biases")


abdulelah_gani = AbdulelahGaniModel(
    abdulelah_gani_p=abdulelah_gani_p,
    abdulelah_gani_s=abdulelah_gani_s,
    abdulelah_gani_t=abdulelah_gani_t,
    properties_contributions=_ag_pc,
    properties_biases=_ag_biases,
)



"""AbdulelahGani Secondary Structures FragmentationModel implementation.

Import and use the AbdulelahGani Secondary Structures FragmentationModel with:

.. code-block:: python

    from ugropy import abdulelah_gani_s

    # Get groups from molecule's name
    tol = abdulelah_gani_s.get_groups("toluene")

    print(tol.subgroups)

    # Get groups from molecule's SMILES
    eth = abdulelah_gani_s.get_groups("CCO", "smiles")

    print(eth.subgroups)

Attributes
----------
abdulelah_gani_s: AbdulelahGaniPSTModel
    AbdulelahGaniPSTModel FragmentationModel :cite:p:`gani`
"""

from ugropy.constants import _csvs
from ugropy.core.frag_classes.abdulelah_gani.abdulelah_gani_pst import (
    AbdulelahGaniPSTModel,
)
from ugropy.models.read_csv import _rd


# =============================================================================
# Abdulelah Gani Secondary Structures FragmentationModel
# =============================================================================
_ag = _csvs / "abdulelah_gani" / "secondary"

_ag_sg = _rd(_ag / "secondary.csv", "group")
_ag_info = _rd(_ag / "info.csv", "group")

abdulelah_gani_s = AbdulelahGaniPSTModel(
    subgroups=_ag_sg,
    subgroups_info=_ag_info,
    allow_overlapping=True,
    allow_free_atoms=True,
)



"""AganiFragmentationResult module."""

import pandas as pd

from rdkit import Chem

from ugropy.core.frag_classes.base.fragmentation_result import (
    FragmentationResult,
)


class AGaniPSTFragmentationResult(FragmentationResult):
    """Abdulelah-Gani group contribution properties estimator.

    This class is to instantiate the Abdulelah-Gani primary, secondary and
    tertiary models.

    Parameters
    ----------
    molecule : Chem.rdchem.Mol
        RDKit molecule object.
    subgroups : dict
        Dictionary of subgroups.
    subgroups_atoms_indexes : dict
        Dictionary of subgroups atoms indexes.
    subgroups_info : pd.DataFrame
        Group's subgroups numbers.

    Attributes
    ----------
    subgroups : dict
        Abdulelah-Gani groups of the molecule.
    subgroups_atoms_indexes : dict
        Abdulelah-Gani groups atoms indexes.
    subgroups_numbers : dict
        Abdulelah-Gani groups numbers.
    """

    def __init__(
        self,
        molecule: Chem.rdchem.Mol,
        subgroups: dict,
        subgroups_atoms_indexes: dict,
        subgroups_info: pd.DataFrame,
    ) -> None:

        super().__init__(molecule, subgroups, subgroups_atoms_indexes)

        self.subgroups_numbers = {}

        if self.subgroups != {}:
            for group, n in self.subgroups.items():
                group_number = int(subgroups_info.loc[group, "group_number"])
                self.subgroups_numbers[group_number] = n
